import base64, copy, argparse, six

def encode_url(url):
    x = six.b(url[url.find('://')+3:])
    return '_url_' + base64.urlsafe_b64encode(x).decode('utf-8').replace('=', '_')

def decode_url(url):
    s = url.replace('_', '=')[5:]
    return base64.urlsafe_b64decode(str(s)).decode('utf-8')

class PackageSource:
    def __init__(self, name=None, url=None, fname=None, recipe=None):
        self.name = name
        self.url = url
        self.fname = fname
        self.recipe = recipe

    def to_name(self):
        return self.name or self.url or self.to_fname()

    def to_fname(self):
        if self.fname is None: self.fname = self.get_encoded_name_url()
        return self.fname

    def get_encoded_name_url(self):
        if self.name is None: return encode_url(self.url)
        else: return self.name.replace('/', '__')

    def get_src_dir(self):
        if self.url.startswith('file://'):
            return self.url[7:] # Remove "file://"
        raise TypeError()


def fname_to_pkg(fname):
    if fname.startswith('_url_'): return PackageSource(name=decode_url(fname), fname=fname)
    else: return PackageSource(name=fname.replace('__', '/'), fname=fname)

class PackageBuild:
    def __init__(self, pkg_src=None, define=None, parent=None, test=False, hash=None, build=None, cmake=None, variant=None, requirements=None, file=None):
        self.pkg_src = pkg_src
        self.define = define or []
        self.parent = parent
        self.test = test
        self.build = build
        self.hash = hash
        self.cmake = cmake
        self.variant = variant or 'Release'
        self.requirements = requirements
        self.file = file

    def merge_defines(self, defines):
        result = copy.copy(self)
        result.define.extend(defines)
        return result

    def merge(self, other):
        result = copy.copy(self)
        if result.define: result.define.extend(other.define)
        else: result.define = other.define
        for field in dir(self):
            if not callable(getattr(self, field)) and not field.startswith("__") and not field in ['define', 'pkg_src']:
                x = getattr(self, field)
                y = getattr(other, field)
                setattr(result, field, y or x)
        return result

    def of(self, parent):
        result = copy.copy(self)
        result.parent = parent.to_fname()
        result.define.extend(parent.define)
        result.variant = parent.variant
        return result

    def to_fname(self):
        if isinstance(self.pkg_src, PackageSource): return self.pkg_src.to_fname()
        else: return self.pkg_src

    def to_name(self):
        if isinstance(self.pkg_src, PackageSource): return self.pkg_src.to_name()
        else: return self.pkg_src

def parse_pkg_build_tokens(args):
    parser = argparse.ArgumentParser()
    parser.add_argument('pkg_src', nargs='?')
    parser.add_argument('-D', '--define', action='append', default=[])
    parser.add_argument('-H', '--hash')
    parser.add_argument('-X', '--cmake')
    parser.add_argument('-f', '--file')
    parser.add_argument('-t', '--test', action='store_true')
    parser.add_argument('-b', '--build', action='store_true')
    return parser.parse_args(args=args, namespace=PackageBuild())




import click, os, sys, shutil, json, six, hashlib, ssl

if sys.version_info[0] < 3:
    try:
        import contextlib
        import lzma
    except:
        try:
            from backports import lzma
        except:
            pass
import tarfile, zipfile

if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess

from six.moves.urllib import request

USE_SYMLINKS=(os.name == 'posix')
USE_CMAKE_TAR=(os.name != 'posix')

__CARBIN_DIR__ = os.path.dirname(os.path.realpath(__file__))

def carbin_dir(*args):
    return os.path.join(__CARBIN_DIR__, *args)

def is_string(obj):
    return isinstance(obj, six.string_types)

def quote(s):
    return json.dumps(s)

class BuildError(Exception):
    def __init__(self, msg=None, data=None):
        self.msg = msg
        self.data = data
    def __str__(self):
        if None: return "Build failed"
        else: return self.msg

def ensure_exists(f):
    if not f:
        raise BuildError("Invalid file path")
    if not os.path.exists(f):
        raise BuildError("File does not exists: " + f)

def can(f):
    try:
        f()
        return True
    except:
        return False

def try_until(*args):
    for arg in args[:-1]:
        try:
            arg()
            return
        except:
            pass
    try:
        args[-1]()
    except:
        raise

def write_to(file, lines):
    content = list((line + "\n" for line in lines))
    if (len(content) > 0):
        with open(file, 'w') as f:
            f.writelines(content)

def mkdir(p):
    if not os.path.exists(p): os.makedirs(p)
    return p

def mkfile(d, file, content, always_write=True):
    mkdir(d)
    p = os.path.join(d, file)
    if not os.path.exists(p) or always_write:
        write_to(p, content)
    return p

def ls(p, predicate=lambda x:True):
    if os.path.exists(p):
        return (d for d in os.listdir(p) if predicate(os.path.join(p, d)))
    else:
        return []

def get_app_dir(*args):
    return os.path.join(click.get_app_dir('carbin'), *args)

def get_cache_path(*args):
    return get_app_dir('cache', *args)

def add_cache_file(key, f):
    mkdir(get_cache_path(key))
    shutil.copy2(f, get_cache_path(key, os.path.basename(f)))

def get_cache_file(key):
    p = get_cache_path(key)
    if os.path.exists(p):
        return os.path.join(p, next(ls(p)))
    else:
        return None

def delete_dir(path):
    if path is not None and os.path.exists(path): shutil.rmtree(path)

def symlink_dir(src, dst):
    for root, dirs, files in os.walk(src):
        all_files = (
            file
            for x in [dirs, files]
            for file in x
            if os.path.islink(os.path.join(root, file)) or os.path.isfile(os.path.join(root, file))
        )
        for file in all_files:
            path = os.path.relpath(root, src)
            d = os.path.join(dst, path)
            mkdir(d)
            relpath = os.path.relpath(os.path.join(root, file), d)
            try:
                os.symlink(relpath, os.path.join(d, file))
            except:
                raise BuildError("Failed to link: {} -> {}".format(os.path.join(root, file), os.path.join(d, file)))

def copy_dir(src, dst):
    for root, dirs, files in os.walk(src):
        for file in files:
            path = os.path.relpath(root, src)
            d = os.path.join(dst, path)
            mkdir(d)
            shutil.copy2(os.path.join(root, file), os.path.join(d, file))

def rm_symlink(file):
    if os.path.islink(file):
        f = os.readlink(file)
        if not os.path.exists(f): os.remove(file)

def rm_symlink_in(file, prefix):
    if os.path.islink(file):
        f = os.readlink(file)
        if not os.path.isabs(f):
            f = os.path.normpath(os.path.join(os.path.dirname(file), f))
        if f.startswith(prefix):
            os.remove(file)

def rm_symlink_dir(d):
    for root, dirs, files in os.walk(d):
        for file in files:
            rm_symlink(os.path.join(root, file))

def rm_symlink_from(d, prefix):
    for root, dirs, files in os.walk(prefix):
        if not root.startswith(d):
            for file in files:
                rm_symlink_in(os.path.join(root, file), d)

def rm_dup_dir(d, prefix, remove_both=True):
    for root, dirs, files in os.walk(d):
        for file in files:
            fullpath = os.path.join(root, file)
            relpath = os.path.relpath(fullpath, d)
            if '..' in relpath:
                raise BuildError('Trying to remove link outside of prefix directory: ' + relpath)
            os.remove(os.path.join(prefix, relpath))
            if remove_both: os.remove(fullpath)

def rm_empty_dirs(d):
    has_files = False
    for x in os.listdir(d):
        p = os.path.join(d, x)
        if os.path.isdir(p) and not os.path.islink(p):
            has_files = has_files or rm_empty_dirs(p)
        else:
            has_files = True
    if not has_files: os.rmdir(d)
    return has_files

def get_dirs(d):
    return (os.path.join(d,o) for o in os.listdir(d) if os.path.isdir(os.path.join(d,o)))

def copy_to(src, dst_dir):
    target = os.path.join(dst_dir, os.path.basename(src))
    if os.path.isfile(src): shutil.copyfile(src, target)
    else: shutil.copytree(src, target)
    return target

def symlink_to(src, dst_dir):
    target = os.path.join(dst_dir, os.path.basename(src))
    os.symlink(src, target)
    return target

class CarbinURLOpener(request.FancyURLopener):
    def http_error_default(self, url, fp, errcode, errmsg, headers):
        if errcode >= 400:
            raise BuildError("Download failed with error {0} for: {1}".format(errcode, url))
        return request.FancyURLopener.http_error_default(self, url, fp, errcode, errmsg, headers)

def download_to(url, download_dir, insecure=False):
    name = url.split('/')[-1]
    file = os.path.join(download_dir, name)
    click.echo("Downloading {0}".format(url))
    bar_len = 1000
    with click.progressbar(length=bar_len, width=70) as bar:
        def hook(count, block_size, total_size):
            percent = int(count*block_size*bar_len/total_size)
            if percent > 0 and percent < bar_len:
                # Hack because we can't set the position
                bar.pos = percent
                bar.update(0)
        context = None
        if insecure: context = ssl._create_unverified_context()
        CarbinURLOpener(context=context).retrieve(url, filename=file, reporthook=hook, data=None)
        bar.update(bar_len)
    if not os.path.exists(file):
        raise BuildError("Download failed for: {0}".format(url))
    return file


def transfer_to(f, dst, copy=False):
    if USE_SYMLINKS and not copy: return symlink_to(f, dst)
    else: return copy_to(f, dst)

def retrieve_url(url, dst, copy=False, insecure=False, hash=None):
    remote = not url.startswith('file://')
    # Retrieve from cache
    if remote and hash:
        f = get_cache_file(hash.replace(':', '-'))
        if f: return f
    f = download_to(url, dst, insecure=insecure) if remote else transfer_to(url[7:], dst, copy=copy)
    if os.path.isfile(f) and hash:
        click.echo("Computing hash: {}".format(hash))
        if check_hash(f, hash):
            if remote: add_cache_file(hash.replace(':', '-'), f)
        else:
            raise BuildError("Hash doesn't match for {0}: {1}".format(url, hash))
    return f

def extract_ar(archive, dst, *kwargs):
    if sys.version_info[0] < 3 and archive.endswith('.xz'):
        with contextlib.closing(lzma.LZMAFile(archive)) as xz:
            with tarfile.open(fileobj=xz, *kwargs) as f:
                f.extractall(dst)
    elif archive.endswith('.zip'):
        with zipfile.ZipFile(archive,'r') as f:
            f.extractall(dst)
    elif tarfile.is_tarfile(archive):
        if USE_CMAKE_TAR:
            cmd([which('cmake'), '-E', 'tar', 'xzf', os.path.abspath(archive)], cwd=dst)
        else:
            tarfile.open(archive, *kwargs).extractall(dst)
    else:
        # Treat as a single source file
        d = os.path.join(dst, 'header')
        mkdir(d)
        copy_to(archive, d)

def hash_file(f, t):
    h = hashlib.new(t)
    h.update(open(f, 'rb').read())
    return h.hexdigest()

def check_hash(f, hash):
    t, h = hash.lower().split(':')
    return hash_file(f, t) == h

def which(p, paths=None, throws=True):
    exes = [p+x for x in ['', '.exe', '.bat']]
    for dirname in list(paths or [])+os.environ['PATH'].split(os.pathsep):
        for exe in exes:
            candidate = os.path.join(os.path.expanduser(dirname), exe)
            if os.path.isfile(candidate):
                return candidate
    if throws: raise BuildError("Can't find file %s" % p)
    else: return None

def merge(*args):
    result = {}
    for d in args:
        result.update(dict(d or {}))
    return result

def flat(*args):
    for arg in args:
        for x in arg:
            for y in x: yield y

def yield_from(f):
    @six.wraps(f)
    def g(*args, **kwargs):
        return flat(f(*args, **kwargs))
    return g

def cmd(args, env=None, capture=None, **kwargs):
    e = merge(os.environ, env)
    c = capture or ''
    stdout = None
    stderr = None
    if c == 'out' or c == 'all': stdout = subprocess.PIPE
    if c == 'err' or c == 'all': stderr = subprocess.PIPE
    child = subprocess.Popen(args, stdout=stdout, stderr=stderr, env=e, **kwargs)
    out = child.communicate()
    if child.returncode != 0:
        raise BuildError(msg='Command failed: ' + str(args), data=e)
    return out

def as_list(x):
    if is_string(x): return [x]
    else: return list(x)

def to_define_dict(xs):
    result = {}
    for x in xs:
        if '=' in x:
            p = x.split('=')
            result[p[0]] = p[1]
        else:
            result[x] = ''
    return result

def as_dict_str(d):
    result = {}
    for x in d:
        result[x] = str(d[x])
    return result

def actual_path(path, start=None):
    if os.path.isabs(path):
        return path
    return os.path.normpath(os.path.join(start or os.getcwd(), os.path.expanduser(path)))

def down_carbin_cmake():
    url = "https://github.com/gottingen/carbin-cmake/archive/master.zip"
    p = os.path.join(os.path.curdir, "carbin/carbin")
    tp = os.path.curdir
    f = retrieve_url(url, p)
    extract_ar(archive = f, dst = p)
    cp = os.path.join(p, "carbin-cmake-master")
    if not os.path.exists(cp):
       print ("download carbin cmake template error")
       exit(1)
    files = os.listdir(cp)
    for file in files:
        if not file == "README.md" and not file == "LICENSE" and not file == ".gitignore":
            print (file)
            copy_to(os.path.join(cp,file), tp)



class Commander:
    def __init__(self, paths=None, env=None, verbose=False):
        self.paths = paths
        self.env = env
        self.verbose = verbose

    def _get_paths_env(self):
        if self.paths is not None:
            return { 'PATH': os.pathsep.join(list(self.paths)+[os.environ['PATH']]) }
        else: return None

    def _cmd(self, name, args=None, options=None, env=None, **kwargs):
        exe = which(name, self.paths)
        option_args = ["{0}={1}".format(key, value) for key, value in six.iteritems(options or {})]
        c = [exe] + option_args + as_list(args or [])
        if self.verbose: click.secho(' '.join(c), bold=True)
        return cmd(c, env=as_dict_str(merge(self.env, self._get_paths_env(), env)), **kwargs)

    def __getattr__(self, name):
        c = name.replace('_', '-')
        def f(*args, **kwargs):
            return self._cmd(c, *args, **kwargs)
        return f

    def __getitem__(self, name):
        def f(*args, **kwargs):
            return self._cmd(name, *args, **kwargs)
        return f

    def __contains__(self, name):
        exe = which(name, self.paths, throws=False)
        return exe is not None



import os, shutil, shlex, six, inspect, click, contextlib, uuid, sys, functools

from carbin.builder import Builder
from carbin.package import fname_to_pkg
from carbin.package import PackageSource
from carbin.package import PackageBuild
from carbin.package import parse_pkg_build_tokens
import carbin.util as util
from carbin.types import returns
from carbin.types import params

__CARBIN_DIR__ = os.path.dirname(os.path.realpath(__file__))
__CARBIN_CMAKE_DIR__ = os.path.join(__CARBIN_DIR__, 'cmake')

@params(s=six.string_types)
def parse_deprecated_alias(s):
    i = s.find(':', 0, max(s.find('://'), s.find(':\\')))
    if i > 0: 
        click.echo("WARNING: Using ':' for aliases is now deprecated.")
        return s[0:i], s[i+1:]
    else: return None, s

@params(s=six.string_types)
def parse_alias(s):
    i = s.find(',')
    if i > 0: return s[0:i], s[i+1:]
    else: return parse_deprecated_alias(s)

@params(s=six.string_types)
def parse_src_name(url, default=None):
    x = url.split('@')
    p = x[0]
    # If the same name is used, then reduce to the same name
    if '/' in p:
        ps = p.split('/')
        if functools.reduce(lambda x, y: x == y, ps):
            p = ps[0]
    v = default
    if len(x) > 1: v = x[1]
    return (p, v)

def cmake_set(var, val, quote=True, cache=None, description=None):
    x = val
    if quote: x = util.quote(val)
    if cache is None or cache.lower() == 'none':
        yield "set({0} {1})".format(var, x)
    else:
        yield 'set({0} {1} CACHE {2} "{3}")'.format(var, x, cache, description or '')

def cmake_append(var, *vals, **kwargs):
    quote = True
    if 'quote' in kwargs: quote = kwargs['quote']
    x = ' '.join(vals)
    if quote: x = ' '.join([util.quote(val) for val in vals])
    yield 'list(APPEND {0} {1})'.format(var, x)

def cmake_if(cond, *args):
    yield 'if ({})'.format(cond)
    for arg in args:
        for line in arg:
            yield '    ' + line
    yield 'endif()'

def cmake_else(*args):
    yield 'else ()'
    for arg in args:
        for line in arg:
            yield '    ' + line

def parse_cmake_var_type(key, value):
    if ':' in key:
        p = key.split(':')
        return (p[0], p[1].upper(), value)
    elif value.lower() in ['on', 'off', 'true', 'false']: 
        return (key, 'BOOL', value)
    else:
        return (key, 'STRING', value)

def find_cmake(p, start):
    if p and not os.path.isabs(p):
        absp = util.actual_path(p, start)
        if os.path.exists(absp): return absp
        else:
            x = util.carbin_dir('cmake', p)
            if os.path.exists(x): return x
            elif os.path.exists(x + '.cmake'): return x + '.cmake'
    return p



PACKAGE_SOURCE_TYPES = (six.string_types, PackageSource, PackageBuild)

class CarbinPrefix:
    def __init__(self, prefix, verbose=False, build_path=None):
        self.prefix = os.path.abspath(prefix or 'carbin')
        self.verbose = verbose
        self.build_path_var = build_path
        self.cmd = util.Commander(paths=[self.get_path('bin')], env=self.get_env(), verbose=self.verbose)
        self.toolchain = self.write_cmake()

    def log(self, *args):
        if self.verbose: click.secho(' '.join([str(arg) for arg in args]), bold=True)

    def check(self, f, *args):
        if self.verbose and not f(*args):
            raise util.BuildError('ASSERTION FAILURE: ', ' '.join([str(arg) for arg in args]))


    def get_env(self):
        return {
            'LD_LIBRARY_PATH': self.get_path('lib'),
            'PKG_CONFIG_PATH': self.pkg_config_path()
        }

    def write_cmake(self, always_write=False, **kwargs):
        return util.mkfile(self.get_private_path(), 'carbin.cmake', self.generate_cmake_toolchain(**kwargs), always_write=always_write)

    @returns(inspect.isgenerator)
    @util.yield_from
    def generate_cmake_toolchain(self, toolchain=None, cc=None, cxx=None, cflags=None, cxxflags=None, ldflags=None, std=None, defines=None):
        set_ = cmake_set
        if_ = cmake_if
        else_ = cmake_else
        append_ = cmake_append
        yield set_('CARBIN_PREFIX', self.prefix)
        yield set_('CMAKE_PREFIX_PATH', self.prefix)
        yield if_('${CMAKE_VERSION} VERSION_LESS "3.6.0"',
            ['include_directories(SYSTEM ${CARBIN_PREFIX}/include)'],
            else_(
                set_('CMAKE_CXX_STANDARD_INCLUDE_DIRECTORIES', '${CARBIN_PREFIX}/include'),
                set_('CMAKE_C_STANDARD_INCLUDE_DIRECTORIES', '${CARBIN_PREFIX}/include')
            )
        )
        if toolchain: yield ['include({})'.format(util.quote(os.path.abspath(toolchain)))]
        yield if_('CMAKE_CROSSCOMPILING',
            append_('CMAKE_FIND_ROOT_PATH', self.prefix)
        )
        yield if_('CMAKE_INSTALL_PREFIX_INITIALIZED_TO_DEFAULT',
            set_('CMAKE_INSTALL_PREFIX', self.prefix)
        )
        if cxx: yield set_('CMAKE_CXX_COMPILER', cxx)
        if cc: yield set_('CMAKE_C_COMPILER', cc)
        if std:
            yield if_('NOT "${CMAKE_CXX_COMPILER_ID}" STREQUAL "MSVC"',
                set_('CMAKE_CXX_STD_FLAG', "-std={}".format(std))
            )
        yield if_('"${CMAKE_CXX_COMPILER_ID}" STREQUAL "MSVC"',
            set_('CMAKE_CXX_ENABLE_PARALLEL_BUILD_FLAG', "/MP".format(std))
        )
        if cflags:
            yield set_('CMAKE_C_FLAGS', "$ENV{{CFLAGS}} ${{CMAKE_C_FLAGS_INIT}} {}".format(cflags or ''), cache='STRING')
        if cxxflags or std:
            yield set_('CMAKE_CXX_FLAGS', "$ENV{{CXXFLAGS}} ${{CMAKE_CXX_FLAGS_INIT}} ${{CMAKE_CXX_STD_FLAG}} {}".format(cxxflags or ''), cache='STRING')
        if ldflags:
            for link_type in ['STATIC', 'SHARED', 'MODULE', 'EXE']:
                yield set_('CMAKE_{}_LINKER_FLAGS'.format(link_type), "$ENV{{LDFLAGS}} {0}".format(ldflags), cache='STRING')
        for dkey in defines or {}:
            name, vtype, value = parse_cmake_var_type(dkey, defines[dkey])
            yield set_(name, value, cache=vtype, quote=(vtype != 'BOOL'))
        yield if_('BUILD_SHARED_LIBS',
            set_('CMAKE_WINDOWS_EXPORT_ALL_SYMBOLS', 'ON', cache='BOOL')
        )
        yield set_('CMAKE_FIND_FRAMEWORK', 'LAST', cache='STRING')
        yield set_('CMAKE_INSTALL_RPATH', '${CARBIN_PREFIX}/lib', cache='STRING')


    def get_path(self, *paths):
        p = os.path.join(self.prefix, *paths)
        return p

    def get_private_path(self, *paths):
        return self.get_path('carbin', *paths)

    def get_public_path(self, *paths):
        return self.get_path('etc', 'carbin', *paths)

    def get_recipe_paths(self):
        return [self.get_public_path('recipes')]

    def get_builder_path(self, *paths):
        if self.build_path_var: return os.path.join(self.build_path_var, *paths)
        else: return self.get_private_path('build', *paths)

    @contextlib.contextmanager
    def create_builder(self, name, tmp=False):
        pre = ''
        if tmp: pre = 'tmp-'
        d = self.get_builder_path(pre + name)
        exists = os.path.exists(d)
        util.mkdir(d)
        yield Builder(self, d, exists)
        if tmp: shutil.rmtree(d, ignore_errors=True)

    def get_package_directory(self, *dirs):
        p = self.get_private_path('pkg', *dirs)
        return p

    def get_unlink_directory(self, *dirs):
        return self.get_private_path('unlink', *dirs)

    def get_deps_directory(self, name, *dirs):
        return self.get_package_directory(name, 'deps', *dirs)

    def get_unlink_deps_directory(self, name, *dirs):
        return self.get_unlink_directory(name, 'deps', *dirs)

    def parse_src_file(self, name, url, start=None):
        f = util.actual_path(url, start)
        self.log('parse_src_file actual_path:', start, f)
        if os.path.exists(f): return PackageSource(name=name, url='file://' + f)
        return None

    def parse_src_recipe(self, name, url):
        p, v = parse_src_name(url)
        for rpath in self.get_recipe_paths():
            rp = os.path.normcase(os.path.join(rpath, p, v or ''))
            if os.path.exists(rp):
                return PackageSource(name=name or p, recipe=rp)
        return None

    def parse_src_github(self, name, url):
        p, v = parse_src_name(url, 'HEAD')
        if '/' in p: url = 'https://github.com/{0}/archive/{1}.tar.gz'.format(p, v)
        else: url = 'https://github.com/{0}/{0}/archive/{1}.tar.gz'.format(p, v)
        if name is None: name = p
        return PackageSource(name=name, url=url)

    @returns(PackageSource)
    @params(pkg=PACKAGE_SOURCE_TYPES)
    def parse_pkg_src(self, pkg, start=None, no_recipe=False):
        if isinstance(pkg, PackageSource): return pkg
        if isinstance(pkg, PackageBuild): return self.parse_pkg_src(pkg.pkg_src, start)
        name, url = parse_alias(pkg)
        self.log('parse_pkg_src:', name, url, pkg)
        if '://' not in url:
            return self.parse_src_file(name, url, start) or \
                (None if no_recipe else self.parse_src_recipe(name, url)) or \
                self.parse_src_github(name, url)
        return PackageSource(name=name, url=url)

    @returns(PackageBuild)
    @params(pkg=PACKAGE_SOURCE_TYPES)
    def parse_pkg_build(self, pkg, start=None, no_recipe=False):
        if isinstance(pkg, PackageBuild): 
            pkg.pkg_src = self.parse_pkg_src(pkg.pkg_src, start, no_recipe)
            if pkg.pkg_src.recipe: pkg = self.from_recipe(pkg.pkg_src.recipe, pkg)
            if pkg.cmake: pkg.cmake = find_cmake(pkg.cmake, start)
            return pkg
        else:
            pkg_src = self.parse_pkg_src(pkg, start, no_recipe)
            if pkg_src.recipe: return self.from_recipe(pkg_src.recipe, pkg_src.name)
            else: return PackageBuild(pkg_src)

    def from_recipe(self, recipe, pkg=None, name=None):
        recipe_pkg = os.path.join(recipe, "package.txt")
        util.ensure_exists(recipe_pkg)
        p = next(iter(self.from_file(recipe_pkg, no_recipe=True)))
        self.check(lambda:p.pkg_src is not None)
        requirements = os.path.join(recipe, "requirements.txt")
        if os.path.exists(requirements): p.requirements = requirements
        p.pkg_src.recipe = None
        # Use original name
        if pkg: p.pkg_src.name = pkg.pkg_src.name
        elif name: p.pkg_src.name = name

        if pkg: return p.merge(pkg)
        else: return p

    def from_file(self, file, url=None, no_recipe=False):
        if file is None:
            return
        if not os.path.exists(file):
            self.log("file not found: " + file)
            return
        start = os.path.dirname(file)
        if url is not None and url.startswith('file://'):
            start = url[7:]
        with open(file) as f:
            self.log("parse file: " + file)
            for line in f.readlines():
                tokens = shlex.split(line, comments=True)
                if len(tokens) > 0: 
                    pb = parse_pkg_build_tokens(tokens)
                    ps = self.from_file(util.actual_path(pb.file, start), no_recipe=no_recipe) if pb.file else [self.parse_pkg_build(pb, start=start, no_recipe=no_recipe)]
                    for p in ps: yield p

    def write_parent(self, pb, track=True):
        if track and pb.parent is not None: util.mkfile(self.get_deps_directory(pb.to_fname()), pb.parent, pb.parent)

    def install_deps(self, pb, d, test=False, test_all=False, generator=None, insecure=False):
        for dependent in self.from_file(pb.requirements or os.path.join(d, 'requirements.txt'), pb.pkg_src.url):
            transient = dependent.test or dependent.build
            testing = test or test_all
            installable = not dependent.test or dependent.test == testing
            if installable: 
                self.install(dependent.of(pb), test_all=test_all, generator=generator, track=not transient, insecure=insecure)

    @returns(six.string_types)
    @params(pb=PACKAGE_SOURCE_TYPES, test=bool, test_all=bool, update=bool, track=bool)
    def install(self, pb, test=False, test_all=False, generator=None, update=False, track=True, insecure=False):
        pb = self.parse_pkg_build(pb)
        pkg_dir = self.get_package_directory(pb.to_fname())
        unlink_dir = self.get_unlink_directory(pb.to_fname())
        install_dir = self.get_package_directory(pb.to_fname(), 'install')

        # If its been unlinked, then link it in
        if os.path.exists(unlink_dir):
            if update: shutil.rmtree(unlink_dir)
            else:
                self.link(pb)
                self.write_parent(pb, track=track)
                return "Linking package {}".format(pb.to_name())
        if os.path.exists(pkg_dir):
            print (pkg_dir) 
            self.write_parent(pb, track=track)
            if update: 
                self.remove(pb)
            else:
                print (pkg_dir) 
                return "Package {} already installed".format(pb.to_name())
        with self.create_builder(uuid.uuid4().hex, tmp=True) as builder:
            # Fetch package
            src_dir = builder.fetch(pb.pkg_src.url, pb.hash, (pb.cmake != None), insecure=insecure)
            # Install any dependencies first
            self.install_deps(pb, src_dir, test=test, test_all=test_all, generator=generator, insecure=insecure)
            # Setup cmake file
            if pb.cmake: 
                target = os.path.join(src_dir, 'CMakeLists.txt')
                if os.path.exists(target):
                    os.rename(target, os.path.join(src_dir, builder.cmake_original_file))
                shutil.copyfile(pb.cmake, target)
            # Configure and build
            print ("install dir")
            print (install_dir)
            builder.configure(src_dir, defines=pb.define, generator=generator, install_prefix=install_dir, test=test, variant=pb.variant)
            builder.build(variant=pb.variant)
            # Run tests if enabled
            if test or test_all: builder.test(variant=pb.variant)
            # Install
            builder.build(target='install', variant=pb.variant)
            if util.USE_SYMLINKS: util.symlink_dir(install_dir, self.prefix)
            else: util.copy_dir(install_dir, self.prefix)
        self.write_parent(pb, track=track)
        return "Successfully installed {}".format(pb.to_name())

    @returns(six.string_types)
    @params(pb=PACKAGE_SOURCE_TYPES)
    def ignore(self, pb):
        pb = self.parse_pkg_build(pb)
        pkg_dir = self.get_package_directory(pb.to_fname())
        # If package doesn't exist
        if not os.path.exists(pkg_dir):
            util.mkfile(pkg_dir, "ignore", "ignore")
            return "Ignore package {}".format(pb.to_name())
        else:
            return "Package {} already installed".format(pb.to_name())

    @params(pb=PACKAGE_SOURCE_TYPES, test=bool)
    def build(self, pb, test=False, target=None, generator=None):
        pb = self.parse_pkg_build(pb)
        src_dir = pb.pkg_src.get_src_dir()
        if os.path.exists(os.path.join(src_dir, 'dev-requirements.txt')):
            pb.requirements = os.path.join(src_dir, 'dev-requirements.txt')
        elif os.path.exists(os.path.join(src_dir, 'requirements.txt')):
            pb.requirements = os.path.join(src_dir, 'requirements.txt')
        with self.create_builder(pb.to_fname()) as builder:
            # Install any dependencies first
            self.install_deps(pb, src_dir, generator=generator, test=test)
            # Configure and build
            if not builder.exists: builder.configure(src_dir, defines=pb.define, generator=generator, test=test, variant=pb.variant)
            builder.build(variant=pb.variant, target=target)
            # Run tests if enabled
            if test: builder.test(variant=pb.variant)

    @params(pb=PACKAGE_SOURCE_TYPES)
    def build_path(self, pb):
        pb = self.parse_pkg_build(pb)
        return self.get_builder_path(pb.to_fname(), 'build')

    @params(pb=PACKAGE_SOURCE_TYPES)
    def build_clean(self, pb):
        pb = self.parse_pkg_build(pb)
        p = self.get_builder_path(pb.to_fname())
        if os.path.exists(p): shutil.rmtree(p)

    @params(pb=PACKAGE_SOURCE_TYPES)
    def build_configure(self, pb):
        pb = self.parse_pkg_build(pb)
        src_dir = pb.pkg_src.get_src_dir()
        if 'ccmake' in self.cmd:
            self.cmd.ccmake([src_dir], cwd=self.build_path(pb))
        elif 'cmake-gui' in self.cmd:
            self.cmd.cmake_gui([src_dir], cwd=self.build_path(pb))

    @params(pkg=PACKAGE_SOURCE_TYPES)
    def remove(self, pkg):
        self.unlink(pkg, delete=True)

    @params(pkg=PACKAGE_SOURCE_TYPES)
    def unlink(self, pkg, delete=False):
        pkg = self.parse_pkg_src(pkg)
        pkg_dir = self.get_package_directory(pkg.to_fname())
        unlink_dir = self.get_unlink_directory(pkg.to_fname())
        self.log("Unlink:", pkg_dir)
        if os.path.exists(pkg_dir):
            if util.USE_SYMLINKS:
                util.rm_symlink_from(os.path.join(pkg_dir, 'install'), self.prefix)
            else:
                util.rm_dup_dir(os.path.join(pkg_dir, 'install'), self.prefix, remove_both=False)
            util.rm_empty_dirs(self.prefix)
            if delete: shutil.rmtree(pkg_dir)
            else:
                util.mkdir(self.get_unlink_directory())
                os.rename(pkg_dir, unlink_dir)

    @params(pkg=PACKAGE_SOURCE_TYPES)
    def link(self, pkg):
        pkg = self.parse_pkg_src(pkg)
        pkg_dir = self.get_package_directory(pkg.to_fname())
        unlink_dir = self.get_unlink_directory(pkg.to_fname())
        if os.path.exists(unlink_dir):
            util.mkdir(self.get_package_directory())
            os.rename(unlink_dir, pkg_dir)
            if util.USE_SYMLINKS: util.symlink_dir(os.path.join(pkg_dir, 'install'), self.prefix)
            else: util.copy_dir(os.path.join(pkg_dir, 'install'), self.prefix)
        # Relink dependencies
        for dep in util.ls(self.get_unlink_directory(), os.path.isdir):
            ls = util.ls(self.get_unlink_deps_directory(dep), os.path.isfile)
            if pkg.to_fname() in ls: self.link(dep)

    def _list_files(self, pkg=None, top=True):
        if pkg is None:
            return util.ls(self.get_package_directory(), os.path.isdir)
        else:
            p = self.parse_pkg_src(pkg)
            ls = util.ls(self.get_deps_directory(p.to_fname()), os.path.isfile)
            if top: return [p.to_fname()]+list(ls)
            else: return ls

    def list(self, pkg=None, recursive=False, top=True):
        for d in self._list_files(pkg, top):
            p = fname_to_pkg(d)
            if os.path.exists(self.get_package_directory(d)): yield p
            if recursive:
                for child in self.list(p, recursive=recursive, top=False):
                    yield child

    def clean(self):
        if util.USE_SYMLINKS:
            util.delete_dir(self.get_private_path())
            util.rm_symlink_dir(self.prefix)
            util.rm_empty_dirs(self.prefix)
        else:
            for p in self.list():
                self.remove(p)
            util.delete_dir(self.get_private_path())

    def clean_cache(self):
        p = util.get_cache_path()
        if os.path.exists(p): shutil.rmtree(util.get_cache_path())

    def pkg_config_path(self):
        libs = []
        for p in ['lib', 'lib64', 'share']:
            libs.append(self.get_path(p, 'pkgconfig'))
        return os.pathsep.join(libs)

    @contextlib.contextmanager
    def try_(self, msg=None, on_fail=None):
        try:
            yield
        except util.BuildError as err:
            if err.msg: click.echo(err.msg)
            if msg: click.echo(msg)
            if on_fail: on_fail()
            if self.verbose: 
                if err.data: click.echo(err.data)
                raise
            sys.exit(1)
        except:
            extype, exvalue, extraceback = sys.exc_info()
            click.echo("Unexpected error: " + str(extype))
            click.echo(str(exvalue))
            if msg: click.echo(msg)
            if on_fail: on_fail()
            if self.verbose: raise
            sys.exit(1)




import click, os, multiprocessing, six

import carbin.util as util

class Builder:
    def __init__(self, prefix, top_dir, exists=False):
        self.prefix = prefix
        self.top_dir = top_dir
        self.build_dir = self.get_path('build')
        self.exists = exists
        self.cmake_original_file = '__carbin_original_cmake_file__.cmake'

    def get_path(self, *args):
        return os.path.join(self.top_dir, *args)

    def get_build_path(self, *args):
        return self.get_path('build', *args)

    def is_make_generator(self):
        return os.path.exists(self.get_build_path('Makefile'))

    def cmake(self, options=None, use_toolchain=False, **kwargs):
        if use_toolchain: return self.prefix.cmd.cmake(options=util.merge({'-DCMAKE_TOOLCHAIN_FILE': self.prefix.toolchain}, options), **kwargs)
        else: return self.prefix.cmd.cmake(options=options, **kwargs)

    def show_log(self, log):
        if self.prefix.verbose and os.path.exists(log):
            click.echo(open(log).read())

    def show_logs(self):
        self.show_log(self.get_build_path('CMakeFiles', 'CMakeOutput.log'))
        self.show_log(self.get_build_path('CMakeFiles', 'CMakeError.log'))

    def targets(self):
        out = None
        try:
            out, err = self.cmake(args=['--build', self.build_dir, '--target', 'help'], capture='out')
        except:
            pass
        for line in (out or '').splitlines():
            if line.startswith(six.b('... ')):
                yield line[4:]

    def fetch(self, url, hash=None, copy=False, insecure=False):
        self.prefix.log("fetch:", url)
        if insecure: url = url.replace('https', 'http')
        f = util.retrieve_url(url, self.top_dir, copy=copy, insecure=insecure, hash=hash)
        if os.path.isfile(f):
            click.echo("Extracting archive {0} ...".format(f))
            util.extract_ar(archive=f, dst=self.top_dir)
        return next(util.get_dirs(self.top_dir))

    def configure(self, src_dir, defines=None, generator=None, install_prefix=None, test=True, variant=None):
        self.prefix.log("configure")
        util.mkdir(self.build_dir)
        args = [
            src_dir, 
            '-DCARBIN_CMAKE_DIR={}'.format(util.carbin_dir('cmake')), 
            '-DCARBIN_CMAKE_ORIGINAL_SOURCE_FILE={}'.format(os.path.join(src_dir, self.cmake_original_file))
        ]
        for d in defines or []:
            args.append('-D{0}'.format(d))
        if generator is not None: args = ['-G', generator] + args
        if self.prefix.verbose: args.extend(['-DCMAKE_VERBOSE_MAKEFILE=On'])
        if test: args.extend(['-DBUILD_TESTING=On'])
        else: args.extend(['-DBUILD_TESTING=Off'])
        args.extend(['-DCMAKE_BUILD_TYPE={}'.format(variant or 'Release')])
        if install_prefix is not None: args.extend(['-DCMAKE_INSTALL_PREFIX=' + install_prefix])
        try:
            self.cmake(args=args, cwd=self.build_dir, use_toolchain=True)
        except:
            self.show_logs()
            raise

    def build(self, target=None, variant=None, cwd=None):
        self.prefix.log("build")
        args = ['--build', self.build_dir]
        if variant is not None: args.extend(['--config', variant])
        if target is not None: args.extend(['--target', target])
        if self.is_make_generator(): 
            args.extend(['--', '-j', str(multiprocessing.cpu_count())])
            if self.prefix.verbose: args.append('VERBOSE=1')
        self.cmake(args=args, cwd=cwd)

    def test(self, variant=None):
        self.prefix.log("test")
        if 'check' in self.targets():
            self.build(target='check', variant=variant or 'Release')
        else:
            self.prefix.cmd.ctest((self.prefix.verbose and ['-VV'] or []) + ['-C', variant], cwd=self.build_dir)



import pytest

import sys, os, tarfile, carbin.util, shutil

from six.moves import shlex_quote

__appveyor__ = 'APPVEYOR' in os.environ
appveyor_skip = pytest.mark.skipif(__appveyor__, reason="Trimmed windows tests for appveyor")

__test_dir__ = os.path.dirname(os.path.realpath(__file__))

__carbin_exe__ = carbin.util.which('carbin')

__has_pkg_config__ = carbin.util.can(lambda: carbin.util.which('pkg-config'))

def get_path(*ps):
    return os.path.join(__test_dir__, *ps)

def get_exists_path(*ps):
    result = get_path(*ps)
    assert os.path.exists(result)
    return result

def get_toolchain(*ps):
    return get_exists_path('toolchains', *ps)

def basename(p):
    d, b = os.path.split(p)
    if len(b) > 0: return b
    else: return basename(d)

def create_ar(archive, src):
    with tarfile.open(archive, mode='w:gz') as f:
        name = basename(src)
        f.add(src, arcname=name)

class DirForTests:
    def __init__(self, tmp_dir):
        self.tmp_dir = tmp_dir

    def dir(self, *args):
        return DirForTests(os.path.join(self.tmp_dir, *args))

    def mkdir(self, *args):
        p = os.path.join(self.tmp_dir, *args)
        carbin.util.mkdir(p)
        return DirForTests(p)

    def cmd(self, *args, **kwargs):
        print(args)
        if 'cwd' not in kwargs: kwargs['cwd'] = self.tmp_dir
        carbin.util.cmd(*args, shell=True, **kwargs)

    def cmds(self, g, **kwargs):
        for x in g:
            print(x)
            if x.startswith('assert_path'):
                p = x[11:].strip()
                self.assert_path(p)
            else:
                self.cmd(x, **kwargs)

    def write_to(self, f, content):
        p = self.get_path(f)
        carbin.util.write_to(p, content)
        return p

    def assert_path(self, *ps):
        p = os.path.join(*ps)
        if not os.path.isabs(p):
            p = os.path.join(self.tmp_dir, p)
        print(p)
        assert os.path.exists(p)

    def get_path(self, *ps):
        return os.path.join(self.tmp_dir, *ps)

@pytest.fixture
def d(tmpdir):
    return DirForTests(tmpdir.strpath)

def remove_empty_elements(xs):
    for x in xs:
        if x is not None:
            s = str(x)
            if len(s) > 0: yield s

class CarbinCmd:
    def __init__(self, prefix=None, build_path=None):
        self.prefix = prefix
        self.build_path = build_path

    def __call__(self, *args):
        p = [__carbin_exe__]
        if self.prefix is not None: p.append('--prefix {}'.format(self.prefix))
        if self.build_path is not None: p.append('--build-path {}'.format(self.build_path))
        return ' '.join(p+list(remove_empty_elements(args)))

def carbin_cmd(*args):
    return CarbinCmd()(*args)

def install_cmds(url, lib=None, alias=None, init=None, remove='remove', list_='list', size=1, recipes=None, prefix=None, build_path=None, variants=None, base_size=0, files=None):
    cg = CarbinCmd(prefix=prefix, build_path=build_path)
    if recipes: base_size = base_size + 1
    n = str(size + base_size)
    yield cg('init', init)
    default_variants = ['--debug', '--release', ''] if not __appveyor__ else ['--release']
    for variant in variants or default_variants:
        yield cg(list_)
        yield cg('clean', '-y')
        yield cg('init', init)
        yield cg(list_)
        yield cg('size', '0')
        if recipes: yield cg('install', recipes)
        yield cg('install', '--verbose --test', variant, url)
        yield cg('size', n)
        for f in files or []:
            yield 'assert_path ' + f
        yield cg(list_)
        if __has_pkg_config__ and lib is not None:
            yield cg('pkg-config', '--list-all')
            yield cg('pkg-config', '--exists', lib)
            yield cg('pkg-config', '--cflags --libs', lib)
        if alias is None: yield cg(remove, '--verbose -y', url)
        else: yield cg(remove, '--verbose -y', alias)
        yield cg('size', str(base_size))
        yield cg(list_)
        if not __appveyor__:
            yield cg('install', '--verbose --test', variant, url)
            yield cg(list_)
            yield cg('size', n)
            yield cg('clean', '-y')
            yield cg(list_)
            yield cg('size', '0')

def build_cmds(url=None, init=None, size=0, defines=None, prefix=None, build_path=None, variants=None):
    cg = CarbinCmd(prefix=prefix, build_path=build_path)
    yield cg('init', init)
    default_variants = ['--debug', '--release', ''] if not __appveyor__ else ['--release']
    for variant in variants or default_variants:
        yield cg('size', '0')
        yield cg('build', '--verbose -C -y', url)
        yield cg('size', '0')
        yield cg('build', '--verbose --test', variant, defines, url)
        yield cg('size', str(size))
        yield cg('build', '--verbose --test', variant, defines, url)
        yield cg('size', str(size))
        yield cg('build', '--verbose --path', url)
        if not __appveyor__:
            yield cg('build', '--verbose --test -C -y', variant, defines, url)
            yield cg('size', str(size))
            yield cg('build', '--verbose --test', variant, defines, url)
            yield cg('size', str(size))
        yield cg('clean', '-y')
        yield cg('size', '0')
        if init: yield cg('init', init)

def test_tar(d):
    ar = d.get_path('libsimple.tar.gz')
    create_ar(archive=ar, src=get_exists_path('libsimple'))
    d.cmds(install_cmds(url=ar, lib='simple'))

@appveyor_skip
def test_tar_alias(d):
    ar = d.get_path('libsimple.tar.gz')
    create_ar(archive=ar, src=get_exists_path('libsimple'))
    d.cmds(install_cmds(url='simple,'+ar, lib='simple', alias='simple'))

def test_dir(d):
    d.cmds(install_cmds(url=get_exists_path('libsimple'), lib='simple'))

def test_relative_dir(d):
    p = os.path.relpath(get_exists_path('libsimple'), d.get_path())
    d.cmds(install_cmds(url=p, lib='simple'))

@pytest.mark.xfail(strict=True)
def test_debug_dir_fail(d):
    d.cmds(install_cmds(url=get_exists_path('libsimpledebug'), lib='simple', variants=['--debug']))

def test_debug_dir_pass(d):
    d.cmds(install_cmds(url=get_exists_path('libsimpledebug'), lib='simple', variants=['--release', '']))

if carbin.util.USE_SYMLINKS:
    def test_local_dir(d):
        carbin.util.copy_to(get_exists_path('libsimple'), d.tmp_dir)
        d.cmds(install_cmds(url='.', lib='simple'), cwd=d.get_path('libsimple'))

def test_dir_custom_build_path(d):
    d.cmds(install_cmds(url=get_exists_path('libsimple'), lib='simple', build_path=d.get_path('my_build')))

def test_recipe_simple(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='simple', lib='simple', recipes=recipes))

def test_recipe_simple_branch(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='simple@master', lib='simple', recipes=recipes))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_recipe_simple_branch_fail(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='simple@nonexistent', lib='simple', recipes=recipes))

@appveyor_skip
def test_recipe_simpleflag(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='simpleflag', recipes=recipes))

def test_recipe_simplemultiflag(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='simplemultiflag', recipes=recipes))

@appveyor_skip
def test_recipe_basicapp(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='basicapp', lib='simple', alias='simple', size=2, recipes=recipes))

@appveyor_skip
def test_recipe_basicappnoreq(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='basicappnoreq', lib='simple', alias='simple', size=2, recipes=recipes))

@appveyor_skip
def test_recipe_simple_x(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='simpleheader', recipes=recipes))

@appveyor_skip
def test_symlink_in_prefix(d):
    p = d.get_path('usr')
    if carbin.util.USE_SYMLINKS:
        carbin.util.mkdir(os.path.join(p, 'share', 'man'))
        os.symlink(os.path.join(p, 'share', 'man'), os.path.join(p, 'man'))
    d.cmds(install_cmds(url=get_exists_path('libsimple'), lib='simple', prefix=p))

def test_prefix(d):
    d.cmds(install_cmds(url=get_exists_path('libsimple'), lib='simple', prefix=d.get_path('usr')))

def test_relative_prefix(d):
    d.cmds(install_cmds(url=get_exists_path('libsimple'), lib='simple', prefix='usr'))

@appveyor_skip
def test_recipe_prefix(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='simple', lib='simple', recipes=recipes, prefix=d.get_path('usr')))

@appveyor_skip
def test_recipe_relative_prefix(d):
    recipes=get_exists_path('basicrecipes') + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    d.cmds(install_cmds(url='simple', lib='simple', recipes=recipes, prefix='usr'))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_xcmake_fail(d):
    d.cmds(install_cmds(url=get_exists_path('libsimplebare'), lib='simple'))

def test_xcmake(d):
    url = get_exists_path('libsimplebare') + ' --cmake ' + get_exists_path('libsimple', 'CMakeLists.txt')
    d.cmds(install_cmds(url=url, lib='simple', alias=get_exists_path('libsimplebare')))

@appveyor_skip
def test_xcmake_s(d):
    url = get_exists_path('libsimplebare') + ' -X ' + get_exists_path('libsimple', 'CMakeLists.txt')
    d.cmds(install_cmds(url=url, lib='simple', alias=get_exists_path('libsimplebare')))

def test_single_header_xcmake(d):
    f = get_exists_path('simpleinclude', 'include', 'simple.h')
    url = f + ' --cmake header'
    d.cmds(install_cmds(url=url, alias=f, files=['carbin/include/simple.h']))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_header_xcmake_fail(d):
    d.cmds(install_cmds(url=get_exists_path('simpleinclude')))

def test_header_xcmake(d):
    url = get_exists_path('simpleinclude') + ' --cmake header'
    d.cmds(install_cmds(url=url, alias=get_exists_path('simpleinclude'), files=['carbin/include/simple.h']))

def test_header_xcmake_include(d):
    url = get_exists_path('simpleheader') + ' --cmake header -DINCLUDE_DIR=inc'
    d.cmds(install_cmds(url=url, alias=get_exists_path('simpleheader'), files=['carbin/include/simple/simple.h']))

def test_header_xcmake_header(d):
    url = get_exists_path('simpleheader') + ' --cmake header -DHEADER_DIR=inc/simple'
    d.cmds(install_cmds(url=url, alias=get_exists_path('simpleheader'), files=['carbin/include/simple/simple.h']))

def test_binary_xcmake(d):
    url = get_exists_path('simpleinclude') + ' --cmake binary'
    d.cmds(install_cmds(url=url, lib='simple', alias=get_exists_path('simpleinclude')))

@appveyor_skip
def test_header_xcmake_dir(d):
    url = get_exists_path('simpleinclude') + ' --cmake ' + get_exists_path('cmake', 'header_cmake_dir.cmake')
    d.cmds(install_cmds(url=url, alias=get_exists_path('simpleinclude')))

def test_xcmake_ref_original(d):
    url = get_exists_path('libsimple') + ' --cmake ' + get_exists_path('cmake', 'use_original_cmake.cmake')
    d.cmds(install_cmds(url=url, lib='simple', alias=get_exists_path('libsimple')))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_xcmake_ref_original_fail(d):
    url = get_exists_path('libsimplebare') + ' --cmake ' + get_exists_path('cmake', 'use_original_cmake.cmake')
    d.cmds(install_cmds(url=url, lib='simple', alias=get_exists_path('libsimplebare')))

@appveyor_skip
def test_rm(d):
    d.cmds(install_cmds(url=get_exists_path('libsimple'), lib='simple', remove='rm'))

@appveyor_skip
def test_ls(d):
    d.cmds(install_cmds(url=get_exists_path('libsimple'), lib='simple', list_='ls'))

@appveyor_skip
def test_update(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('simpleapp')),
        carbin_cmd('size', '1'),
        carbin_cmd('install', '--verbose --test --update', 'app,'+get_exists_path('simpleapp')),
        carbin_cmd('size', '1'),
        carbin_cmd('rm', '--verbose -y', 'app'),
        carbin_cmd('size', '0')
    ])

def test_update_reqs(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('simpleapp')),
        carbin_cmd('size', '1'),
        carbin_cmd('install', '--verbose --test --update', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y', 'app'),
        carbin_cmd('size', '1')
    ])


@appveyor_skip
def test_rm_with_symlink(d):
    p = d.get_path('usr')
    share = os.path.join(p, 'share')
    carbin.util.mkfile(share, 'data.txt', 'data')
    if carbin.util.USE_SYMLINKS:
        os.symlink(os.path.join(share, 'data.txt'), os.path.join(share, 'data.sym'))
    else:
        carbin.util.mkfile(share, 'data.sym', 'data')
    os.remove(os.path.join(share, 'data.txt'))
    cg = CarbinCmd(prefix=p)
    d.cmds([
        cg('install', '--verbose --test', 'simple,'+get_exists_path('libsimple')),
        cg('size', '1'),
        cg('rm', '--verbose -y', 'simple'),
        cg('size', '0')
    ])
    assert os.path.exists(os.path.join(share))
    assert not os.path.exists(os.path.join(share, 'data.txt'))
    assert os.path.lexists(os.path.join(share, 'data.sym'))

@appveyor_skip
def test_rm_all(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test --update', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y --all'),
        carbin_cmd('size', '0')
    ])

def test_unlink1(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', 'simple,'+get_exists_path('libsimple')),
        carbin_cmd('size', '1'),
        carbin_cmd('rm', '--verbose -y --unlink', 'simple'),
        carbin_cmd('size', '0'),
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y', 'app'),
        carbin_cmd('size', '1'),
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2')
    ])

@appveyor_skip
def test_unlink2(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y --unlink', 'simple'),
        carbin_cmd('size', '0'),
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y', 'app'),
        carbin_cmd('size', '1'),
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2')
    ])

@appveyor_skip
def test_unlink3(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y --unlink', 'app'),
        carbin_cmd('size', '1'),
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y', 'app'),
        carbin_cmd('size', '1'),
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2')
    ])

@appveyor_skip
def test_unlink4(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', 'simple,'+get_exists_path('libsimple')),
        carbin_cmd('size', '1'),
        carbin_cmd('rm', '--verbose -y --unlink', 'simple'),
        carbin_cmd('size', '0'),
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y', 'simple'),
        carbin_cmd('size', '0'),
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2')
    ])

@appveyor_skip
def test_unlink_update(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', 'app,'+get_exists_path('simpleapp')),
        carbin_cmd('size', '1'),
        carbin_cmd('rm', '--verbose -y --unlink', 'app'),
        carbin_cmd('install', '--verbose --test --update', 'app,'+get_exists_path('basicapp')),
        carbin_cmd('size', '2'),
        carbin_cmd('rm', '--verbose -y', 'app'),
        carbin_cmd('size', '1')
    ])

@appveyor_skip
def test_unlink_shared(d):
    d.cmds([
        carbin_cmd('init', '--shared'),
        carbin_cmd('install', '--verbose --test', get_exists_path('libsimple2')),
        carbin_cmd('size', '1')
    ])
    libname = None
    if sys.platform == 'darwin':
        libname = 'libsimple.dylib'
    elif sys.platform.startswith('linux'):
        libname = 'libsimple.so'
    if libname:
        d.assert_path('carbin', 'lib', libname)
    d.cmds([
        carbin_cmd('rm', '--verbose -y --unlink', get_exists_path('libsimple2')),
        carbin_cmd('install', '--verbose', get_exists_path('libsimple2')),
        carbin_cmd('size', '1')
    ])
    if libname:
        d.assert_path('carbin', 'lib', libname)

@appveyor_skip
def test_unlink_pkg(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', get_exists_path('libsimple')),
        carbin_cmd('size', '1'),
        carbin_cmd('rm', '--verbose -y --unlink', get_exists_path('libsimple')),
    ])
    shutil.rmtree(d.get_path('carbin', 'carbin', 'pkg'))
    d.cmds([
        carbin_cmd('install', '--verbose --test', get_exists_path('libsimple')),
        carbin_cmd('size', '1')
    ])


@appveyor_skip
def test_build_dir(d):
    d.cmds(build_cmds(get_exists_path('libsimple')))

@appveyor_skip
def test_build_dir_notest(d):
    d.cmds([
        carbin_cmd('build', '--verbose', get_exists_path('buildtesting')),
    ])

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_build_dir_with_test(d):
    d.cmds([
        carbin_cmd('build', '--verbose --test', get_exists_path('buildtesting')),
    ])

@appveyor_skip
def test_build_dir_dev(d):
    d.cmds(build_cmds(get_exists_path('basicapp-dev'), size=1))

def test_build_relative_dir(d):
    p = os.path.relpath(get_exists_path('libsimple'), d.get_path())
    d.cmds(build_cmds(p))

def test_build_dir_custom_build_path(d):
    d.cmds(build_cmds(get_exists_path('libsimple'), build_path=d.get_path('my_build')))

@appveyor_skip
def test_build_dir_custom_build_path2(d):
    d.cmds(build_cmds(get_exists_path('basicapp'), build_path=d.get_path('my_build'), size=1))

@appveyor_skip
def test_build_relative_dir_custom_build_path(d):
    p = os.path.relpath(get_exists_path('libsimple'), d.get_path())
    d.cmds(build_cmds(p, build_path=d.get_path('my_build')))

def test_build_relative_dir_custom_build_path2(d):
    p = os.path.relpath(get_exists_path('basicapp'), d.get_path())
    d.cmds(build_cmds(p, build_path=d.get_path('my_build'), size=1))

@appveyor_skip
def test_tmp_build_dir_custom_build_path(d):
    d.mkdir('tmp').cmds(build_cmds(get_exists_path('libsimple'), build_path=d.get_path('my_build')))

def test_tmp_build_dir_custom_build_path2(d):
    d.mkdir('tmp').cmds(build_cmds(get_exists_path('basicapp'), build_path=d.get_path('my_build'), size=1))

@appveyor_skip
def test_tmp_relative_build_dir_custom_build_path(d):
    p = os.path.relpath(get_exists_path('libsimple'), d.get_path('tmp'))
    d.mkdir('tmp').cmds(build_cmds(p, build_path=d.get_path('my_build')))

def test_tmp_relative_build_dir_custom_build_path2(d):
    p = os.path.relpath(get_exists_path('basicapp'), d.get_path('tmp'))
    d.mkdir('tmp').cmds(build_cmds(p, build_path=d.get_path('my_build'), size=1))

def test_build_current_dir(d):
    cwd = get_exists_path('libsimple')
    d.cmds(build_cmds(prefix=d.get_path('carbin')), cwd=cwd)

def test_build_target(d):
    d.cmds([
        carbin_cmd('build', '--verbose', '--target simpleapp', get_exists_path('simpleapp')),
        carbin_cmd('build', '--verbose', '--target simpleapptest', get_exists_path('simpleapp')),
        carbin_cmd('size', '0')
    ])

def test_build_install_target(d):
    # d.mkdir('app')
    d.mkdir('tmp').cmds([
        carbin_cmd('build', '--verbose', '--target install', '-DCMAKE_INSTALL_PREFIX="' + d.get_path('app') + '"', get_exists_path('basicapp')),
        carbin_cmd('size', '1')
    ])


@pytest.mark.xfail(strict=True)
def test_build_target_fail(d):
    d.cmds([carbin_cmd('build', '--verbose', '--target xyz', get_exists_path('simpleapp'))])

@appveyor_skip
def test_dir_deprecated_alias(d):
    d.cmds(install_cmds(url='simple,'+get_exists_path('libsimple'), lib='simple', alias='simple'))

@appveyor_skip
def test_dir_alias(d):
    d.cmds(install_cmds(url='simple,'+get_exists_path('libsimple'), lib='simple', alias='simple'))

def test_init_cpp11(d):
    d.cmds([
        carbin_cmd('init', '--std=c++0x'),
        carbin_cmd('install', '--verbose --test', get_exists_path('libsimple11'))
    ])

@appveyor_skip
def test_reqs_alias_file(d):
    reqs_file = d.write_to('reqs', [shlex_quote('simple,'+get_exists_path('libsimple'))])
    d.cmds(install_cmds(url='--file {}'.format(reqs_file), lib='simple', alias='simple'))

@appveyor_skip
def test_reqs_file(d):
    reqs_file = d.write_to('reqs', [shlex_quote(get_exists_path('libsimple'))])
    d.cmds(install_cmds(url='--file {}'.format(reqs_file), lib='simple', alias=get_exists_path('libsimple')))

def test_reqs_alias_f(d):
    reqs_file = d.write_to('reqs', [shlex_quote('simple,'+get_exists_path('libsimple'))])
    d.cmds(install_cmds(url='-f {}'.format(reqs_file), lib='simple', alias='simple'))

@appveyor_skip
def test_reqs_file_flag(d):
    reqs_file = d.write_to('reqs', ['-f'+shlex_quote(get_exists_path('basicapp', 'requirements.txt'))])
    d.cmds(install_cmds(url='--file {}'.format(reqs_file), lib='simple', alias='simple'))

@appveyor_skip
def test_reqs_f(d):
    reqs_file = d.write_to('reqs', [shlex_quote(get_exists_path('libsimple'))])
    d.cmds(install_cmds(url='-f {}'.format(reqs_file), lib='simple', alias=get_exists_path('libsimple')))

def test_reqs_f2(d):
    reqs_file = get_exists_path('basicapp', 'requirements.txt')
    d.cmds(install_cmds(url='-f {}'.format(reqs_file), lib='simple', alias='simple'))

@appveyor_skip
def test_reqs_f3(d):
    reqs_file = d.write_to('reqs', [shlex_quote(get_exists_path('libsimple'))])
    d.write_to('requirements.txt', [shlex_quote(get_exists_path('basicapp'))])
    d.cmds(install_cmds(url='-f {}'.format(reqs_file), lib='simple', alias=get_exists_path('libsimple')))

def test_reqs_hash(d):
    ar = d.get_path('libsimple.tar.gz')
    create_ar(archive=ar, src=get_exists_path('libsimple'))
    h = carbin.util.hash_file(ar, 'sha1')
    reqs_file = d.write_to('reqs', ["{0} --hash=sha1:{1}".format(shlex_quote(ar), h)])
    d.cmds(install_cmds(url='--file {}'.format(reqs_file), lib='simple', alias=ar))

@appveyor_skip
def test_without_reqs_f(d):
    d.write_to('requirements.txt', [shlex_quote(get_exists_path('libsimple'))])
    d.cmds(install_cmds(url='', lib='simple', alias=get_exists_path('libsimple')))

@appveyor_skip
def test_without_reqs_f_dev(d):
    d.write_to('dev-requirements.txt', [shlex_quote(get_exists_path('libsimple'))])
    d.cmds(install_cmds(url='', lib='simple', alias=get_exists_path('libsimple')))

@appveyor_skip
def test_without_reqs_f_dev2(d):
    d.write_to('dev-requirements.txt', [shlex_quote(get_exists_path('libsimple'))])
    d.write_to('requirements.txt', ['fail'])
    d.cmds(install_cmds(url='', lib='simple', alias=get_exists_path('libsimple')))

@appveyor_skip
def test_without_reqs_f2(d):
    d.write_to('requirements.txt', [shlex_quote(get_exists_path('basicapp'))])
    d.cmds(install_cmds(url=get_exists_path('libsimple'), lib='simple'))


@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_reqs_hash_fail(d):
    ar = d.get_path('libsimple.tar.gz')
    create_ar(archive=ar, src=get_exists_path('libsimple'))
    h = 'xxx'
    reqs_file = d.write_to('reqs', ["{0} --hash=sha1:{1}".format(shlex_quote(ar), h)])
    d.cmds(install_cmds(url='--file {}'.format(reqs_file), lib='simple', alias=ar))

def test_reqs_recipe(d):
    recipes=shlex_quote(get_exists_path('basicrecipes')) + ' -DCARBIN_TEST_DIR="' + __test_dir__ + '"'
    reqs_file = d.write_to('reqs', [recipes, 'simple'])
    d.cmds(install_cmds(url='--file {}'.format(reqs_file), lib='simple', alias='simple', base_size=1))

@appveyor_skip
def test_app_include_dir(d):
    d.cmds(install_cmds(url=get_exists_path('basicapp-include'), lib='simple', alias='simple', size=2))

@appveyor_skip
def test_app_header_dep(d):
    d.cmds(install_cmds(url=get_exists_path('basicapp-simpleinclude'), alias='simple', size=2))

@appveyor_skip
# @pytest.mark.xfail(strict=True)
def test_app_header_werror(d):
    d.cmds(install_cmds(url=get_exists_path('basicapp-simpleinclude-werror'), alias='simple', size=2))

# Basic app needs pkg-config
if __has_pkg_config__:

    @appveyor_skip
    def test_app_dir(d):
        d.cmds(install_cmds(url=get_exists_path('basicapp'), lib='simple', alias='simple', size=2))

    @appveyor_skip
    def test_app_indirect_dir(d):
        d.cmds(install_cmds(url=get_exists_path('basicapp-indirect'), lib='simple', alias='simple', size=2))

    @appveyor_skip
    def test_xapp_dir(d):
        d.cmds(install_cmds(url=get_exists_path('basicappx'), lib='simple', alias='simple', size=2))

    def test_appdebug_dir_pass1(d):
        d.cmds(install_cmds(url=get_exists_path('basicappdebug'), lib='simple', alias='simple', size=2, variants=['--release']))

    @appveyor_skip
    def test_appdebug_dir_pass2(d):
        d.cmds(install_cmds(url=get_exists_path('basicappdebug'), lib='simple', alias='simple', size=2, variants=['']))

    @appveyor_skip
    @pytest.mark.xfail(strict=True)
    def test_appdebug_dir_fail(d):
        d.cmds(install_cmds(url=get_exists_path('basicappdebug'), lib='simple', alias='simple', size=2, variants=['--debug']))

    @appveyor_skip
    def test_build_flag_child(d):
        d.cmds([
            carbin_cmd('install', '--verbose', get_exists_path('basicappbuild')),
            carbin_cmd('list'),
            carbin_cmd('size', '2'),
            carbin_cmd('remove', '-y simple'),
            carbin_cmd('list'),
            carbin_cmd('size', '1')
        ])

    def test_build_flag_parent(d):
        d.cmds([
            carbin_cmd('install', '--verbose', get_exists_path('basicappbuild')),
            carbin_cmd('list'),
            carbin_cmd('size', '2'),
            carbin_cmd('remove', '-y', get_exists_path('basicappbuild')),
            carbin_cmd('list'),
            carbin_cmd('size', '1')
        ])


    def test_build_app_dir(d):
        d.cmds(build_cmds(get_exists_path('basicapp'), size=1))

@appveyor_skip
def test_install_simple_app_test_with_test(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', get_exists_path('simpleapptest')),
        carbin_cmd('list'),
        carbin_cmd('size', '2'),
        carbin_cmd('remove', '-y simple'),
        carbin_cmd('list'),
        carbin_cmd('size', '1')
    ])

@appveyor_skip
def test_install_simple_app_test_with_test_all(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test-all', get_exists_path('simpleapptest')),
        carbin_cmd('list'),
        carbin_cmd('size', '2'),
        carbin_cmd('remove', '-y simple'),
        carbin_cmd('list'),
        carbin_cmd('size', '1')
    ])

@appveyor_skip
def test_install_simple_app_test_all_with_test_with_all(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test-all', get_exists_path('simpleapptestall')),
        carbin_cmd('list'),
        carbin_cmd('size', '3'),
        carbin_cmd('remove', '-y simple'),
        carbin_cmd('list'),
        carbin_cmd('size', '2')
    ])

@appveyor_skip
def test_install_simple_app_test_all_with_test(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', get_exists_path('simpleapptestall')),
        carbin_cmd('list'),
        carbin_cmd('size', '2'),
        carbin_cmd('remove', '-y simpleapptest'),
        carbin_cmd('list'),
        carbin_cmd('size', '1')
    ])

@appveyor_skip
def test_install_simple_app_test_all_without_test(d):
    d.cmds([
        carbin_cmd('install', '--verbose', get_exists_path('simpleapptestall')),
        carbin_cmd('list'),
        carbin_cmd('size', '1'),
        carbin_cmd('remove', '-y', get_exists_path('simpleapptestall')),
        carbin_cmd('list'),
        carbin_cmd('size', '0')
    ])

@appveyor_skip
def test_install_simple_app_test_without_test(d):
    d.cmds([
        carbin_cmd('install', '--verbose', get_exists_path('simpleapptest')),
        carbin_cmd('list'),
        carbin_cmd('size', '1')
    ])

@appveyor_skip
def test_build_simple_app_test_with_test(d):
    d.cmds(build_cmds(get_exists_path('simpleapptest'), size=1))

@appveyor_skip
def test_build_simple_app_test_without_test(d):
    d.cmds([
        carbin_cmd('build', '--verbose', get_exists_path('simpleapptest')),
        carbin_cmd('size', '0')
    ])

@appveyor_skip
def test_install_simple_basic_app_test_with_test_all(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test-all', get_exists_path('simplebasicapptest')),
        carbin_cmd('list'),
        carbin_cmd('size', '3'),
        carbin_cmd('remove', '-y simple'),
        carbin_cmd('list'),
        carbin_cmd('size', '1')
    ])

@appveyor_skip
def test_install_simple_basic_app_test_with_test(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', get_exists_path('simplebasicapptest')),
        carbin_cmd('list'),
        carbin_cmd('size', '3'),
        carbin_cmd('remove', '-y simple'),
        carbin_cmd('list'),
        carbin_cmd('size', '1')
    ])

@appveyor_skip
def test_install_simple_basic_app_test_without_test(d):
    d.cmds([
        carbin_cmd('install', '--verbose ', get_exists_path('simplebasicapptest')),
        carbin_cmd('list'),
        carbin_cmd('size', '1'),
        carbin_cmd('remove', '-y', get_exists_path('simplebasicapptest')),
        carbin_cmd('list'),
        carbin_cmd('size', '0')
    ])


@appveyor_skip
def test_install_simple_basic_app2_test_with_test_all(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test-all', get_exists_path('simplebasicapp')),
        carbin_cmd('list'),
        carbin_cmd('size', '4'),
        carbin_cmd('remove', '-y simple'),
        carbin_cmd('list'),
        carbin_cmd('size', '2')
    ])

@appveyor_skip
def test_install_simple_basic_app2_test_with_test(d):
    d.cmds([
        carbin_cmd('install', '--verbose --test', get_exists_path('simplebasicapp')),
        carbin_cmd('list'),
        carbin_cmd('size', '2'),
        carbin_cmd('remove', '-y simplebasicapptest'),
        carbin_cmd('list'),
        carbin_cmd('size', '1')
    ])

@appveyor_skip
def test_install_simple_basic_app2_test_without_test(d):
    d.cmds([
        carbin_cmd('install', '--verbose ', get_exists_path('simplebasicapp')),
        carbin_cmd('list'),
        carbin_cmd('size', '1'),
        carbin_cmd('remove', '-y', get_exists_path('simplebasicapp')),
        carbin_cmd('list'),
        carbin_cmd('size', '0')
    ])

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_flags_fail(d):
    d.cmds([carbin_cmd('install', '--verbose --test -DCARBIN_FLAG=Off', get_path('libsimpleflag'))])

def test_flags(d):
    p = get_exists_path('libsimpleflag')
    d.cmds(install_cmds(url='-DCARBIN_FLAG=On {}'.format(p), alias=p))

@appveyor_skip
def test_flags_init(d):
    d.cmds(install_cmds(init='-DCARBIN_FLAG=On', url=get_exists_path('libsimpleflag')))

def test_build_flags(d):
    d.cmds(build_cmds(get_exists_path('libsimpleflag'), defines='-DCARBIN_FLAG=On'))

@appveyor_skip
def test_build_flags_init(d):
    d.cmds(build_cmds(init='-DCARBIN_FLAG=On', url=get_exists_path('libsimpleflag')))

@appveyor_skip
def test_flags_init_integer(d):
    d.cmds(install_cmds(init='-DCARBIN_FLAG=1', url=get_exists_path('libsimpleflag')))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_flags_fail_integer(d):
    d.cmds([carbin_cmd('install --verbose --test -DCARBIN_FLAG=0', get_path('libsimpleflag'))])

@appveyor_skip
def test_flags_integer(d):
    p = get_exists_path('libsimpleflag')
    d.cmds(install_cmds(url='-DCARBIN_FLAG=1 {}'.format(p), alias=p))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_flags_fail_define(d):
    d.cmds([carbin_cmd('install', '--verbose --test --define CARBIN_FLAG=Off', get_path('libsimpleflag'))])

def test_flags_define(d):
    d.cmds([carbin_cmd('install', '--verbose --test --define CARBIN_FLAG=On', get_exists_path('libsimpleflag'))])

@appveyor_skip
def test_flags_toolchain(d):
    d.cmds([
        carbin_cmd('init', '--toolchain', get_toolchain('toolchainflag.cmake')),
        carbin_cmd('install', '--verbose --test', get_exists_path('libsimpleflag'))
    ])

def test_flags_toolchain_prefix(d):
    cg = CarbinCmd(d.get_path('usr'))
    d.cmds([
        cg('init', '--toolchain', get_toolchain('toolchainflag.cmake')),
        cg('install', '--verbose --test', get_exists_path('libsimpleflag'))
    ])

@appveyor_skip
def test_flags_reqs_f(d):
    p = get_exists_path('libsimpleflag')
    reqs_file = d.write_to('reqs', [shlex_quote(p) + ' -DCARBIN_FLAG=On'])
    d.cmds(install_cmds(url='-f {}'.format(reqs_file), alias=p))

@appveyor_skip
def test_flags_reqs_f2(d):
    p = get_exists_path('libsimpleflag')
    reqs_file = d.write_to('reqs', [shlex_quote(p)])
    d.cmds(install_cmds(url='-f {} -DCARBIN_FLAG=On'.format(reqs_file), alias=p))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_flags_reqs_f2_fail(d):
    p = get_exists_path('libsimpleflag')
    reqs_file = d.write_to('reqs', [shlex_quote(p)])
    d.cmds(install_cmds(url='-f {} -DCARBIN_FLAG=Off'.format(reqs_file), alias=p))

def test_multiflags(d):
    p = get_exists_path('libsimplemultiflag')
    d.cmds(install_cmds(url='-DCARBIN_FLAG1=On -DCARBIN_FLAG2=On {}'.format(p), alias=p))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_multiflags_fail1(d):
    d.cmds([carbin_cmd('install', '--verbose --test -DCARBIN_FLAG1=Off -DCARBIN_FLAG2=Off', get_path('libsimplemultiflag'))])

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_multiflags_fail2(d):
    d.cmds([carbin_cmd('install', '--verbose --test -DCARBIN_FLAG1=Off -DCARBIN_FLAG2=On', get_path('libsimplemultiflag'))])

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_multiflags_fail3(d):
    d.cmds([carbin_cmd('install', '--verbose --test -DCARBIN_FLAG1=On', get_path('libsimplemultiflag'))])

@appveyor_skip
def test_comments_reqs_f(d):
    p = get_exists_path('libsimple')
    reqs_file = d.write_to('reqs', [shlex_quote(p) + ' #A comment', '# Another comment'])
    d.cmds(install_cmds(url='-f {}'.format(reqs_file), alias=p))

def test_shared_init(d):
    d.cmds(install_cmds(init='--shared', url=get_exists_path('libsimple'), lib='simple'))

@appveyor_skip
def test_static_init(d):
    d.cmds(install_cmds(init='--static', url=get_exists_path('libsimple'), lib='simple'))

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_shared_static_init(d):
    d.cmds(install_cmds(init='--shared --static', url=get_exists_path('libsimple'), lib='simple'))

def test_ignore(d):
    d.cmds([
        carbin_cmd('size', '0'),
        carbin_cmd('ignore', '--verbose', get_exists_path('basicapp')),
        carbin_cmd('size', '1'),
        carbin_cmd('install', '--verbose', get_exists_path('basicapp')),
        carbin_cmd('size', '1'),
        carbin_cmd('remove', '--verbose -y', get_exists_path('basicapp')),
        carbin_cmd('size', '0')
    ])

@appveyor_skip
@pytest.mark.xfail(strict=True)
def test_ignore_dep(d):
    d.cmds([
        carbin_cmd('ignore', '--verbose simple'),
        carbin_cmd('install', '--verbose', get_exists_path('basicapp'))
    ])

@appveyor_skip
def test_symlink_dir(d):
    d.cmds([
        carbin_cmd('install', get_path('symlinkdir'))
    ])
    assert os.path.exists(d.get_path('carbin', 'data', 'sdir', 'file.txt'))

def test_cmake_trouble(d):
    d.cmds([
        carbin_cmd('install', get_exists_path('cmake-trouble')),
        carbin_cmd('install', get_exists_path('libsimple'))
    ])

def test_subdir(d):
    d.cmds(install_cmds(url='-X subdir {}'.format(get_exists_path('libsimplesubdir')), lib='simple', alias=get_exists_path('libsimplesubdir')))



import click, os, functools, sys

from carbin import __version__
from carbin.prefix import CarbinPrefix
from carbin.prefix import PackageBuild
import carbin.util as util


aliases = {
    'rm': 'remove',
    'ls': 'list'
}

class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        if cmd_name in aliases:
            return click.Group.get_command(self, ctx, aliases[cmd_name])
        return None


@click.group(cls=AliasedGroup, context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(version=__version__, prog_name='carbin')
@click.option('-p', '--prefix', envvar='CARBIN_PREFIX', help='Set prefix used to install packages')
@click.option('-v', '--verbose', is_flag=True, envvar='VERBOSE', help="Enable verbose mode")
@click.option('-B', '--build-path', envvar='CARBIN_BUILD_PATH', help='Set the path for the build directory to use when building the package')
@click.pass_context
def cli(ctx, prefix, verbose, build_path):
    ctx.obj = {}
    if prefix: ctx.obj['PREFIX'] = prefix
    if verbose: ctx.obj['VERBOSE'] = verbose
    if build_path: ctx.obj['BUILD_PATH'] = build_path

def use_prefix(f):
    @click.option('-p', '--prefix', help='Set prefix used to install packages')
    @click.option('-v', '--verbose', is_flag=True, help="Enable verbose mode")
    @click.option('-B', '--build-path', help='Set the path for the build directory to use when building the package')
    @click.pass_obj
    @functools.wraps(f)
    def w(obj, prefix, verbose, build_path, *args, **kwargs):
        p = CarbinPrefix(prefix or obj.get('PREFIX'), verbose or obj.get('VERBOSE'), build_path or obj.get('BUILD_PATH'))
        f(p, *args, **kwargs)
    return w

@cli.command(name='init')
@use_prefix
@click.option('-t', '--toolchain', required=False, help="Set cmake toolchain file to use")
@click.option('--cc', required=False, help="Set c compiler")
@click.option('--cxx', required=False, help="Set c++ compiler")
@click.option('--cflags', required=False, help="Set additional c flags")
@click.option('--cxxflags', required=False, help="Set additional c++ flags")
@click.option('--ldflags', required=False, help="Set additional linker flags")
@click.option('--std', required=False, help="Set C++ standard if available")
@click.option('-D', '--define', multiple=True, help="Extra configuration variables to pass to CMake")
@click.option('--shared', is_flag=True, help="Set toolchain to build shared libraries by default")
@click.option('--static', is_flag=True, help="Set toolchain to build static libraries by default")
def init_command(prefix, toolchain, cc, cxx, cflags, cxxflags, ldflags, std, define, shared, static):
    """ Initialize install directory """
    if shared and static:
        click.echo("ERROR: shared and static are not supported together")
        sys.exit(1)
    defines = util.to_define_dict(define)
    if shared: defines['BUILD_SHARED_LIBS'] = 'On'
    if static: defines['BUILD_SHARED_LIBS'] = 'Off'
    prefix.write_cmake(
        always_write=True, 
        toolchain=toolchain, 
        cc=cc,
        cxx=cxx,
        cflags=cflags, 
        cxxflags=cxxflags, 
        ldflags=ldflags, 
        std=std, 
        defines=defines)
    util.down_carbin_cmake()

@cli.command(name='install')
@use_prefix
@click.option('-U', '--update', is_flag=True, help="Update package")
@click.option('-t', '--test', is_flag=True, help="Test package before installing by running ctest or check target")
@click.option('--test-all', is_flag=True, help="Test all packages including its dependencies before installing by running ctest or check target")
@click.option('-f', '--file', default=None, help="Install packages listed in the file")
@click.option('-D', '--define', multiple=True, help="Extra configuration variables to pass to CMake")
@click.option('-G', '--generator', envvar='CARBIN_GENERATOR', help='Set the generator for CMake to use')
@click.option('-X', '--cmake', help='Set cmake file to use to build project')
@click.option('--debug', is_flag=True, help="Install debug version")
@click.option('--release', is_flag=True, help="Install release version")
@click.option('--insecure', is_flag=True, help="Don't use https urls")
@click.argument('pkgs', nargs=-1, type=click.STRING)
def install_command(prefix, pkgs, define, file, test, test_all, update, generator, cmake, debug, release, insecure):
    """ Install packages """
    if debug and release:
        click.echo("ERROR: debug and release are not supported together")
        sys.exit(1)
    variant = 'Release'
    if debug: variant = 'Debug'
    if not file and not pkgs:
        if os.path.exists('dev-requirements.txt'): file = 'dev-requirements.txt'
        else: file = 'requirements.txt'
    pbs = [PackageBuild(pkg, cmake=cmake, variant=variant) for pkg in pkgs]
    for pbu in util.flat([prefix.from_file(file), pbs]):
        pb = pbu.merge_defines(define)
        with prefix.try_("Failed to build package {}".format(pb.to_name()), on_fail=lambda: prefix.remove(pb)):
            click.echo(prefix.install(pb, test=test, test_all=test_all, update=update, generator=generator, insecure=insecure))

@cli.command(name='ignore')
@use_prefix
@click.argument('pkgs', nargs=-1, type=click.STRING)
def ignore_command(prefix, pkgs):
    """ Ignore packages """
    pbs = [PackageBuild(pkg) for pkg in pkgs]
    for pb in pbs:
        with prefix.try_("Failed to ignore package {}".format(pb.to_name()), on_fail=lambda: prefix.remove(pb)):
            click.echo(prefix.ignore(pb))

@cli.command(name='build')
@use_prefix
@click.option('-t', '--test', is_flag=True, help="Test package by running ctest or check target")
@click.option('-c', '--configure', is_flag=True, help="Configure cmake")
@click.option('-C', '--clean', is_flag=True, help="Remove build directory")
@click.option('-P', '--path', is_flag=True, help="Show path to build directory")
@click.option('-D', '--define', multiple=True, help="Extra configuration variables to pass to CMake")
@click.option('-T', '--target', default=None, help="Cmake target to build")
@click.option('-y', '--yes', is_flag=True, default=False)
@click.option('-G', '--generator', envvar='CARBIN_GENERATOR', help='Set the generator for CMake to use')
@click.option('--debug', is_flag=True, help="Build debug version")
@click.option('--release', is_flag=True, help="Build release version")
@click.argument('pkg', nargs=1, default='.', type=click.STRING)
def build_command(prefix, pkg, define, test, configure, clean, path, yes, target, generator, debug, release):
    """ Build package """
    pb = PackageBuild(pkg).merge_defines(define)
    if debug and release:
        click.echo("ERROR: debug and release are not supported together")
        sys.exit(1)
    pb.variant = 'Release'
    if debug: pb.variant = 'Debug'
    with prefix.try_("Failed to build package {}".format(pb.to_name())):
        if configure: prefix.build_configure(pb)
        elif path: click.echo(prefix.build_path(pb))
        elif clean: 
            if not yes: yes = click.confirm("Are you sure you want to delete the build directory?")
            if yes: prefix.build_clean(pb)
        else: prefix.build(pb, test=test, target=target, generator=generator)

@cli.command(name='remove')
@use_prefix
@click.argument('pkgs', nargs=-1, type=click.STRING)
@click.option('-y', '--yes', is_flag=True, default=False)
@click.option('-U', '--unlink', is_flag=True, default=False, help="Unlink package but don't remove it")
@click.option('-A', '--all', is_flag=True, default=False, help="Select all packages installed")
def remove_command(prefix, pkgs, yes, unlink, all):
    """ Remove packages """
    if all: pkgs = [None]
    verb = "unlink" if unlink else "remove"
    pkgs_set = set((dep.name for pkg in pkgs for dep in prefix.list(pkg, recursive=True)))
    click.echo("The following packages will be removed:")
    for pkg in pkgs_set: click.echo(pkg)
    if not yes: yes = click.confirm("Are you sure you want to {} these packages?".format(verb))
    if yes:
        for pkg in pkgs_set:
            with prefix.try_("Failed to {} package {}".format(verb, pkg)):
                prefix.unlink(pkg, delete=not unlink)
                click.echo("{} package {}".format(verb, pkg))

@cli.command(name='list')
@use_prefix
def list_command(prefix):
    """ List installed packages """
    for pkg in prefix.list():
        click.echo(pkg.name)

# TODO: Make this command hidden
@cli.command(name='size')
@use_prefix
@click.argument('n')
def size_command(prefix, n):
    pkgs = len(list(util.ls(prefix.get_package_directory(), os.path.isdir)))
    if pkgs != int(n):
        raise util.BuildError("Not the correct number of items: {}".format(pkgs))

@cli.command(name='clean')
@use_prefix
@click.option('-y', '--yes', is_flag=True, default=False)
@click.option('--cache', is_flag=True, default=False, help="Removes any cache files")
def clean_command(prefix, yes, cache):
    """ Clear directory """
    if cache:
        prefix.clean_cache()
    else:
        if not yes: yes = click.confirm("Are you sure you want to delete all carbin packages in {}?".format(prefix.prefix))
        if yes: prefix.clean()

@cli.command(name='pkg-config', context_settings=dict(
    ignore_unknown_options=True,
))
@use_prefix
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
def pkg_config_command(prefix, args):
    """ Pkg config """
    prefix.cmd.pkg_config(args)


if __name__ == '__main__':
    cli()




from setuptools import setup, find_packages
import os, re, sys

def get_version(package):
    """Return package version as listed in `__version__` in `init.py`."""
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)

def get_requires(filename):
    requirements = []
    with open(filename) as req_file:
        for line in req_file.read().splitlines():
            if not line.strip().startswith("#"):
                requirements.append(line)
    return requirements

project_requirements = get_requires("requirements.txt")

if os.name == 'posix' and sys.version_info[0] < 3: project_requirements.append('subprocess32')

setup(
    name="carbin",
    version=get_version("carbin"),
    url='https://github.com/gottingen/carbin',
    license='bsd',
    description='Cmake bazel package retrieval',
    author='lothar',
    author_email='lijippy@163.com',
    packages=find_packages(),
    package_data={'cmake': ['*.cmake']},
    install_requires=project_requirements,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'carbin = carbin.cli:cli',
        ]
    },
    zip_safe=False
)



import six, inspect, os

def decorator_with_args(decorator):
    def f(*args, **kwargs):
        def g(func):
            return decorator(func, *args, **kwargs)
        return g
    return f

def identity_decorator(f):
    return f

def is_iterable(obj):
    return hasattr(obj, '__iter__')

def default_checker(obj, _type_):
    return isinstance(obj, _type_), "type '{}'".format(_type_.__name__)

def callable_checker(obj, f):
    return f(obj), f.__name__

def any_checkers(checkers):
    return any(map(lambda e:e[0], checkers))

def format_checkers(checkers):
    failed_checkers = filter(lambda e:not e[0], checkers)
    s_checkers = list(map(lambda e:e[1], failed_checkers))
    return "({})".format(','.join(s_checkers))

def get_checker(x):
    if inspect.isclass(x): return default_checker
    if six.callable(x): return callable_checker
    if is_iterable(x):
        def checker(obj, _type_):
            checkers = map(lambda e:get_checker(e)(obj, e), _type_)
            return any_checkers(checkers), format_checkers(checkers)
        return checker

def require_type(obj, _type_, fname, name=None):
    b, msg = get_checker(_type_)(obj, _type_)
    if not b:
        s = 'Return'
        if name is not None: s = "Parameter '{0}'".format(name)
        raise TypeError("{0} should be {1} but found type '{2}' instead for {3}".format(s, msg, type(obj), fname))
    return obj

DEBUG='DEBUG' in os.environ and len(os.environ['DEBUG']) > 0

if DEBUG:
    @decorator_with_args
    def params(f, **argument_types):
        @six.wraps(f)
        def check_call(*args, **kwargs):
            callargs = inspect.getcallargs(f, *args, **kwargs)
            for name, _type_ in six.iteritems(argument_types):
                require_type(callargs[name], _type_, f.__name__, name)
            return f(*args, **kwargs)
        return check_call

    @decorator_with_args
    def returns(f, _type_):
        @six.wraps(f)
        def check_call(*args, **kwargs):
            return require_type(f(*args, **kwargs), _type_, f.__name__)
        return check_call

else:
    def params(*arg, **kwargs):
        return identity_decorator
    def returns(*arg, **kwargs):
        return identity_decorator
